from click.testing import CliRunner
from parboil.parboil import boil

def test_general():
	runner = CliRunner()

	# help message
	result = runner.invoke(boil, ['list', '--help'])
	assert result.exit_code == 0
	assert 'Usage: boil list' in result.output

def test_list_1():
	runner = CliRunner()

	# normal use
	result = runner.invoke(boil, ['list'])
	assert result.exit_code == 0

	# wrong tpldir
	result = runner.invoke(boil, ['--tpldir', 'SOME/UNKNOWN/DIR', 'list'])
	assert result.exit_code == 1
	assert 'Template folder does not exist.' in result.output

	# custom tpldir
	result = runner.invoke(boil, ['--tpldir', 'test/repo', 'list'])
	assert result.exit_code == 0
	assert 'Listing templates in test/repo' in result.output

	# missing config file
	result = runner.invoke(boil, ['-c', 'SOME/UNKNOWN/FILE.json', 'list'])
	assert result.exit_code == 2
	assert 'No such file or directory' in result.output

	# custom config file
	result = runner.invoke(boil, ['-c', 'test/config.json', 'list'])
	assert result.exit_code == 0
	assert 'Listing templates in test/repo' in result.output




# -*- coding: utf-8 -*-

from pathlib import Path

import pytest

from parboil.project import Project, Repository

TPLDIR = 'test/repo'

def test_repo():
	repo = Repository(TPLDIR)
	assert str(repo.root) == TPLDIR

	for tpl in repo:
		assert tpl in ['pbt', 'test']

	for tpl in repo.projects():
		assert type(tpl) is Project
		assert tpl.name in ['pbt', 'test']



from click.testing import CliRunner
from parboil.parboil import boil
from parboil.version import __version__

def test_boil():
	runner = CliRunner()
	result = runner.invoke(boil)
	assert result.exit_code == 0

	# Version information
	result = runner.invoke(boil, ['--version'])
	assert result.exit_code == 0
	assert __version__ in result.output

	# Unknwon command
	result = runner.invoke(boil, ['unknown-command'])
	assert result.exit_code == 2
	assert 'No such command \'unknown-command\''

	# Missing command
	result = runner.invoke(boil, ['--tpldir', 'SOME/DIR'])
	assert result.exit_code == 2
	assert 'Missing command' in result.output



from click.testing import CliRunner
from parboil.parboil import boil

def test_boil_use():
	runner = CliRunner()

	# use without args not allowed
	result = runner.invoke(boil, ['use'])
	assert result.exit_code == 2

	# help message
	result = runner.invoke(boil, ['use', '--help'])
	assert result.exit_code == 0
	assert 'Usage: boil use' in result.output



# -*- coding: utf-8 -*-

import unicodedata, re
from datetime import datetime
from functools import update_wrapper

from click import pass_context

from jinja2 import nodes
from jinja2.ext import Extension


def pass_tpldir(f):
    @pass_context
    def new_func(ctx, *args, **kwargs):
        if ctx.obj and "TPLDIR" in ctx.obj:
            return ctx.invoke(f, ctx.obj["TPLDIR"], *args, **kwargs)
        else:
            return ctx.invoke(f, ctx.obj["TPLDIR"], *args, **kwargs)

    return update_wrapper(new_func, f)


class JinjaTimeExtension(Extension):
    """
    Adds a {% time %} tag to jinja2

    The argument gets passed to time.strftime.
    """

    tags = {"time"}

    def __init__(self, environment):
        super(JinjaTimeExtension, self).__init__(environment)

        environment.extend(datetime_format="%Y-%m-%d")

    def _time(self, datetime_format):
        if datetime_format is None:
            datetime_format = self.environment.datetime_format

        return datetime.now().strftime(datetime_format)

    def parse(self, parser):
        lineno = next(parser.stream).lineno

        arg = parser.parse_expression()

        call_method = self.call_method(
            "_time",
            [arg],
            lineno=lineno,
        )

        return nodes.Output([call_method], lineno=lineno)


def jinja_filter_fileify(s):
    """
    Django util.text.get_valid_filename
    """
    s = str(s).strip().replace(" ", "_")
    return re.sub(r"(?u)[^-\w.]", "", s)


def jinja_filter_slugify(value, allow_unicode=False):
    """
    Django util.text.slugify
    """
    value = str(value)
    if allow_unicode:
        value = unicodedata.normalize("NFKC", value)
    else:
        value = (
            unicodedata.normalize("NFKD", value)
            .encode("ascii", "ignore")
            .decode("ascii")
        )
    value = re.sub(r"[^\w\s-]", "", value.lower())
    return re.sub(r"[-\s]+", "-", value).strip("-_")



# -*- coding: utf-8 -*-

from pathlib import Path

import pytest
from click.testing import CliRunner

from parboil.parboil import boil
from parboil.project import Project


GLOABL_CONFIG = {'TPLDIR': 'test/repo'}

def test_project_files():
	prj = Project('test')
	prj.setup(GLOABL_CONFIG)

	for file in ['rename_me.txt', 'aaa/rename_me.txt']:
		assert Path(file) in prj.templates
	for file in ['a.txt', 'b.txt', 'c.txt']:
		assert Path(file) in prj.includes

	prj.load()
	assert 'files' in prj.config
	assert 'created' in prj.meta
	assert 'source_type' in prj.meta and prj.meta['source_type'] == 'local'
	assert 'Project' in prj.fields and prj.fields['Project'] == 'Test'

def test_project_errors():
	prj = Project('MISSING')
	with pytest.raises(AttributeError) as e:
		prj.setup({})
		assert 'PARBOIL: key TPLDIR is mandatory in config' in e.message

	prj.setup(GLOABL_CONFIG)

	assert not prj.templates
	assert not prj.includes

	with pytest.raises(FileNotFoundError) as e:
		prj.load(GLOABL_CONFIG)
		assert 'PARBOIL: Project file for template MISSING not found.' in e.message



import shutil, json
from pathlib import Path
from click.testing import CliRunner
from parboil.parboil import boil, PRJ_FILE, META_FILE


TEST_REPO = Path('test/repo')
TEST_OUT  = Path('test/out')
TEST_TPL  = Path('test/templates')


def test_boil_install():
	runner = CliRunner()

	# use without args not allowed
	result = runner.invoke(boil, ['install'])
	assert result.exit_code == 2

	# help message
	result = runner.invoke(boil, ['install', '--help'])
	assert result.exit_code == 0
	assert 'Usage: boil install' in result.output


def test_local():
	"""Test install from local filesystem"""
	runner = CliRunner()

	local_name = 'test'
	local_path = (TEST_TPL / local_name).resolve()
	local_install = Path(local_name)

	install_name = 'tpl'
	install_path = Path(install_name)


	with runner.isolated_filesystem():
		# Install with default name
		result = runner.invoke(boil, ['--tpldir', '.', 'install', str(local_path)])
		assert result.exit_code == 0
		assert f'Installed template {local_name}' in result.output
		assert local_install.is_dir()
		assert (local_install / PRJ_FILE).is_file()
		assert (local_install / META_FILE).is_file()

		with open(local_install / META_FILE) as meta:
			first_run = json.load(meta)['created']

		# Overwrite
		result = runner.invoke(boil, ['--tpldir', '.', 'install', str(local_path)], input="y")
		assert result.exit_code == 0
		assert f'Overwrite existing template named {local_name}' in result.output
		assert f'Removed template {local_name}' in result.output
		assert f'Installed template {local_name}' in result.output
		assert local_install.is_dir()
		assert (local_install / PRJ_FILE).is_file()
		assert (local_install / META_FILE).is_file()

		with open(local_install / META_FILE) as meta:
			second_run = json.load(meta)['created']

		# check metadata
		assert first_run < second_run

		# Install and overwrite from name
		result = runner.invoke(boil, ['--tpldir', '.', 'install', str(local_path), install_name])
		assert result.exit_code == 0
		assert f'Installed template {install_name}' in result.output
		assert install_path.is_dir()
		assert (install_path / PRJ_FILE).is_file()
		assert (install_path / META_FILE).is_file()


def test_github():
	runner = CliRunner()

	install_name = 'pbt'
	repo_name = 'jneug/parboil-template'
	repo_url  = f'https://github.com/{repo_name}'

	install_path = Path(install_name)

	with runner.isolated_filesystem():
		# Install from url
		result = runner.invoke(boil, ['--tpldir', '.', 'install', repo_url, install_name])
		assert result.exit_code == 0
		assert 'Installed template pbt' in result.output
		assert install_path.is_dir()
		assert (install_path / PRJ_FILE).is_file()
		assert (install_path / META_FILE).is_file()

		# Install and overwrite from name
		result = runner.invoke(boil, ['--tpldir', '.', 'install', '--force', '-d', repo_name, install_name])
		assert result.exit_code == 0
		assert 'Removed template pbt' in result.output
		assert 'Installed template pbt' in result.output
		assert install_path.is_dir()
		assert (install_path / PRJ_FILE).is_file()
		assert (install_path / META_FILE).is_file()



# -*- coding: utf-8 -*-

import os
import re
import subprocess
import json
import shutil
import time
import tempfile
from pathlib import Path

import click
from jinja2 import Environment, FileSystemLoader, ChoiceLoader, PrefixLoader

import parboil.fields as fields
from .ext import JinjaTimeExtension, jinja_filter_fileify, jinja_filter_slugify


PRJ_FILE = "project.json"
META_FILE = ".parboil"


class Project(object):
    def __init__(self, name, repository):
        self._name = name
        if type(repository) is Repository:
            self._repo = repository
            self._root_dir = (repository.root / self._name).resolve()
        else:
            self._repo = None
            self._root_dir = (Path(repository) / self._name).resolve()

    @property
    def name(self):
        return self._name

    @property
    def root(self):
        return self._root_dir

    def exists(self):
        return self._root_dir.is_dir()

    def is_project(self):
        return self.exists() and (self._root_dir / PRJ_FILE).is_file()

    def setup(self, load_project=False):
        # setup config files and paths
        self.project_file = self._root_dir / PRJ_FILE
        self.meta_file = self._root_dir / META_FILE
        self.templates_dir = self._root_dir / "template"
        self.includes_dir = self._root_dir / "includes"

        self.meta = dict()
        self.files = dict()
        self.fields = dict()
        self.variables = dict()
        self.templates = list()
        self.includes = list()

        self.templates = list()
        for root, dirs, files in os.walk(self.templates_dir):
            root = Path(root).resolve()
            for name in files:
                dirname = root.relative_to(self.templates_dir)
                self.templates.append(dirname / name)

        self.includes = list()
        for root, dirs, files in os.walk(self.includes_dir):
            root = Path(root).resolve()
            for name in files:
                dirname = root.relative_to(self.includes_dir)
                self.includes.append(dirname / name)

        if load_project:
            self.load()

    def load(self):
        """Loads the project file and some metadata"""

        if not self.project_file:
            self.setup()

        ## project file
        config = dict()
        if self.project_file.exists():
            with open(self.project_file) as f:
                config = json.load(f)
                if "files" not in config:
                    self.files = dict()
                else:
                    self.files = config["files"]
        else:
            raise ProjectFileNotFoundError("Project file not found.")

        if "fields" in config:
            for k, v in config["fields"].items():
                if type(v) is not dict:
                    self.fields[k] = dict(type="default", default=v)
                else:
                    self.fields[k] = v

        ## metafile
        if self.meta_file.is_file():
            with open(self.meta_file) as f:
                self.meta = {**self.meta, **json.load(f)}

    def fill(self, prefilled=dict()):
        # Read user input (if necessary)
        for key, descr in self.fields.items():
            value = None
            if key in prefilled:
                value = prefilled[key]

            if type(descr) is dict and "type" in descr:
                if descr["type"] == "project":
                    subproject = Project(descr["name"], self.root.parent)
                    subproject.setup(load_project=True)
                    subproject.fill({**prefilled, **self.variables})
                    self.templates.append(subproject)
                else:
                    field_callable = f'field_{descr["type"]}'
                    del descr["type"]

                    if hasattr(fields, field_callable):
                        self.variables[key] = getattr(fields, field_callable)(
                            key=key, **descr, value=value, project=self
                        )

    def compile(self, target_dir, jinja=None):
        if not jinja:
            jinja = Environment(
                loader=ChoiceLoader(
                    [
                        FileSystemLoader(self.templates_dir),
                        PrefixLoader(
                            {"includes": FileSystemLoader(self.includes_dir)},
                            delimiter=":",
                        ),
                    ]
                ),
                extensions=[JinjaTimeExtension],
            )
            jinja.filters["fileify"] = jinja_filter_fileify
            jinja.filters["slugify"] = jinja_filter_slugify

        target_dir = Path(target_dir).resolve()

        result = (list(), list())
        for file in self.templates:
            if type(file) is Project:
                for result in file.compile(target_dir):
                    yield result
            else:
                if type(file) is str:
                    file_in = (
                        Path(file[9:]) if file.startswith("includes:") else Path(file)
                    )
                else:
                    file_in = file
                file_out = Path(file_in)
                if str(file_in) in self.files:
                    if type(self.files[str(file_in)]) is str:
                        file_out = self.files[str(file_in)]

                rel_path = file_in.parent
                abs_path = target_dir / rel_path

                # Set some dynamic values
                boil_vars = dict(
                    TPLNAME=self._name,
                    RELDIR="" if rel_path.name == "" else str(rel_path),
                    ABSDIR=str(abs_path),
                    OUTDIR=str(target_dir),
                )

                path_render = jinja.from_string(str(file_out)).render(
                    **self.variables, BOIL=boil_vars
                )

                boil_vars["FILENAME"] = Path(path_render).name
                boil_vars["FILEPATH"] = path_render

                # Render template
                tpl_render = jinja.get_template(str(file)).render(
                    **self.variables, BOIL=boil_vars
                )

                path_render_abs = target_dir / path_render
                if len(tpl_render) > 0:
                    if not path_render_abs.parent.exists():
                        path_render_abs.parent.mkdir(parents=True)

                    with open(path_render_abs, "w") as f:
                        f.write(tpl_render)

                    yield (True, str(file), path_render)
                else:
                    yield (False, str(file), "")

    def save(self):
        """Saves the current meta file to disk"""
        if self.meta_file:
            with open(self.meta_file, "w") as f:
                json.dump(self.meta, f)

    def update(self, hard=False):
        """Update the template from its original source"""
        if not self.meta_file.exists():
            raise ProjectFileNotFoundError(
                "Template metafile does not exist. Can't read update information."
            )

        if self.meta["source_type"] == "github":
            git = subprocess.Popen(["git", "pull", "--rebase"], cwd=self._root_dir)
            git.wait(30)
        elif self.meta["source_type"] == "local":
            if Path(self.meta["source"]).is_dir():
                shutil.rmtree(self._root_dir)
                shutil.copytree(self.meta["source"], self._root_dir)
            else:
                raise ProjectError("Original source directory no longer exists.")
        else:
            raise ProjectError("No source information found.")

        # Update meta file for later updates
        self.meta["updated"] = time.time()
        self.save()

        self.setup(load_project=True)


class Repository(object):
    def __init__(self, root):
        self._root = Path(root)
        self.load()

    @property
    def root(self):
        return self._root

    def exists(self):
        return self._root.is_dir()

    def load(self):
        self._projects = list()
        for child in self._root.iterdir():
            if child.is_dir():
                project_file = child / PRJ_FILE
                if project_file.is_file():
                    self._projects.append(child.name)

    def __len__(self):
        return len(self._projects)

    def __iter__(self):
        yield from self._projects

    def is_installed(self, template):
        tpl_dir = self._root / template
        return tpl_dir.is_dir()
        # return Project(template, self).exists()

    def projects(self):
        for prj in self._projects:
            yield self.get_project(prj)

    def get_project(self, template):
        return Project(template, self)

    def install_from_directory(self, template, source, hard=False, is_repo=False):
        """IF source contains a valid project template it is installed
        into this local repository and the Project object is returned.
        """
        if self.is_installed(template):
            if not hard:
                raise ProjectExistsError(
                    "The template already exists. Delete first or retry install with hard=True."
                )
            else:
                self._delete(template)

        # check source directory
        source = Path(source).resolve()
        if not source.is_dir():
            raise ProjectFileNotFoundError("Source does not exist.")

        if not is_repo:
            project_file = source / PRJ_FILE
            template_dir = source / "template"

            if not project_file.is_file():
                raise ProjectFileNotFoundError(
                    f"The source does not contain a {PRJ_FILE} file."
                )

            if not template_dir.is_dir():
                raise ProjectFileNotFoundError(
                    "The source does not contain a template directory."
                )

            project = self.get_project(template)

            # copy files
            shutil.copytree(source, project.root)

            # create meta file
            project.setup()
            project.meta = {
                "created": time.time(),
                "source_type": "local",
                "source": str(source),
            }
            project.save()

            self.load()
            return project
        else:
            projects = list()

            for child in source.iterdir():
                if child.is_dir():
                    project_file = child / PRJ_FILE
                    if project_file.is_file():
                        try:
                            project = self.install_from_directory(
                                child.name, child, hard=hard
                            )
                            projects.append(project)
                        except:
                            pass

            self.load()
            return projects

    def install_from_github(self, template, url, hard=False, is_repo=False):
        if not is_repo:
            # check target dir
            if self.is_installed(template):
                if not hard:
                    raise ProjectExistsError(
                        "The template already exists. Delete first or retry install with hard=True."
                    )
                else:
                    self._delete(template)

            project = self.get_project(template)

            # do git clone
            # TODO: Does this work on windows?
            git = subprocess.Popen(["git", "clone", url, str(project.root)])
            git.wait(30)

            # create meta file
            project.setup()
            project.meta = {
                "created": time.time(),
                "source_type": "github",
                "source": url,
            }
            project.save()

            self.load()
            return project
        else:
            projects = list()  # return list of installed projects

            # do git clone into temp folder
            with tempfile.TemporaryDirectory() as temp_repo:
                git = subprocess.Popen(["git", "clone", url, temp_repo])
                git.wait(30)

                for child in Path(temp_repo).iterdir():
                    if child.is_dir():
                        project_file = child / PRJ_FILE
                        if project_file.is_file():
                            try:
                                project = self.install_from_directory(
                                    child.name, child, hard=hard
                                )
                                # remove source data
                                del project.meta["source_type"]
                                del project.meta["source"]
                                project.save()

                                projects.append(project)
                            except:
                                pass

            self.load()
            return projects

    def uninstall(self, template):
        self._delete(template)

    def _delete(self, template):
        """Delete a project template from this repository."""
        tpl_dir = self._root / template
        if tpl_dir.is_dir():
            if tpl_dir.is_symlink():
                tpl_dir.unlink()
            else:
                shutil.rmtree(tpl_dir)

    def _reload(self):
        diff = list()
        for child in self._root.iterdir():
            if child.is_dir():
                project_file = child / PRJ_FILE
                if project_file.is_file():
                    if not child.name in self._projects:
                        diff.append(child.name)
                        self._projects.append(child.name)
        self._projects.sort()
        return diff


class ProjectError(Exception):
    pass


class ProjectFileNotFoundError(FileNotFoundError):
    pass


class ProjectExistsError(FileExistsError):
    pass



# -*- coding: utf-8 -*-


"""setup.py: setuptools control."""


from setuptools import setup

setup(
    install_requires=["click", "colorama", "jinja2"],
    entry_points={"console_scripts": ["boil = parboil.parboil:boil"]},
)



# -*- coding: utf-8 -*-


"""
Parboil lets you generate boilerplate projects from template files.

Run boil --help for more info.
"""

import os
import re
import subprocess
import json
import shutil
import time
from pathlib import Path

import click
from colorama import Fore, Back, Style
from jinja2 import Environment, FileSystemLoader, ChoiceLoader, PrefixLoader

import parboil.console as console

from .version import __version__
from .project import (
    Project,
    Repository,
    ProjectError,
    ProjectFileNotFoundError,
    ProjectExistsError,
)
from .ext import (
    pass_tpldir,
    JinjaTimeExtension,
    jinja_filter_fileify,
    jinja_filter_slugify,
)


# set global defaults
CFG_FILE = "config.json"

# TODO: use click.get_app_dir to be plattform independent
CFG_DIR = Path.home() / ".config" / "parboil"


# TODO: Options for debug/verbosity and colors
@click.group()
@click.version_option(version=__version__, prog_name="parboil")
@click.option(
    "-c",
    "--config",
    type=click.File(),
    envvar="BOIL_CONFIG",
    help="Provides a different json config file for this run. Read from stdin with -.",
)
@click.option(
    "--tpldir",
    type=click.Path(file_okay=False, dir_okay=True),
    envvar="BOIL_TPLDIR",
    help="Location of the local template repository.",
)
@click.pass_context
def boil(ctx, config, tpldir):
    ctx.ensure_object(dict)

    # Set default values
    ctx.obj["TPLDIR"] = CFG_DIR / "templates"

    # Load config file
    if config:
        user_cfg = json.load(config)
        ctx.obj = {**ctx.obj, **user_cfg}
    else:
        cfg_file = CFG_DIR / CFG_FILE
        if cfg_file.exists():
            with open(cfg_file) as f:
                cmd_cfg = json.load(f)
                ctx.obj = {**ctx.obj, **cmd_cfg}

    if tpldir:
        ctx.obj["TPLDIR"] = tpldir
    ctx.obj["TPLDIR"] = Path(ctx.obj["TPLDIR"])


@boil.command()
@click.option("-p", "--plain", is_flag=True)
@pass_tpldir
def list(TPLDIR, plain):
    repo = Repository(TPLDIR)
    if repo.exists():
        if len(repo) > 0:
            if plain:
                for project in repo:
                    click.echo(project)
            else:
                console.info(
                    f"Listing templates in {Style.BRIGHT}{TPLDIR}{Style.RESET_ALL}."
                )
                print()
                console.indent(f'⎪ {"name":^12} ⎪ {"created":^24} ⎪ {"updated":^24} ⎪')
                console.indent(f'|{"-"*14}⎪{"-"*26}⎪{"-"*26}|')
                for project in repo.projects():
                    project.setup(load_project=True)

                    meta = dict()
                    if "updated" not in project.meta:
                        meta["updated"] = "never"
                    else:
                        meta["updated"] = time.ctime(int(project.meta["updated"]))
                    if "created" not in project.meta:
                        meta["created"] = "unknown"
                    else:
                        meta["created"] = time.ctime(int(project.meta["created"]))

                    console.indent(
                        f'| {Fore.CYAN}{project.name:<12}{Style.RESET_ALL} | {meta["created"]:>24} | {meta["updated"]:>24} |'
                    )
                print()
        else:
            console.info("No templates installed yet.")
    else:
        console.warn("Template folder does not exist.")
        exit(1)


@boil.command()
@click.option(
    "-f",
    "--force",
    is_flag=True,
    help="Set this flag to overwrite existing templates named TEMPLATE without prompting.",
)
@click.option(
    "-d",
    "--download",
    is_flag=True,
    help="Set this flag if SOURCE is a github repository to download instead of a local directory.",
)
@click.option("-r", "--repo", "is_repo", is_flag=True)
@click.argument("source")
@click.argument("template", required=False)
@click.pass_context
def install(ctx, source, template, force, download, is_repo):
    """
    Install a project template named TEMPLATE from SOURCE to the local template repository.

    SOURCE may be a local directory or the url of a GitHub repository. You may also pass in
    the name of a repository in the form user/repo, but need to set the -d flag to indicate
    it isn't a local directory.
    """
    # TODO: validate templates!
    TPLDIR = ctx.obj["TPLDIR"]
    repo = Repository(TPLDIR)

    # is github url? Than assume -d
    if re.match(r"https?://(www\.)?github\.com", source):
        download = True
    # set missing arguments
    if download:
        if re.match("[A-Za-z_-]+/[A-Za-z_-]+", source):
            source = f"https://github.com/{source}"
        if not template:
            template = source.split("/")[-1]
    else:
        if not template:
            template = Path(source).name

    if not is_repo and not force and repo.is_installed(template):
        if not console.question(
            f"Overwrite existing template named {Fore.CYAN}{template}{Style.RESET_ALL}",
            color=Fore.YELLOW,
            echo=click.confirm,
        ):
            ctx.abort()

    try:
        if download:
            projects = repo.install_from_github(
                template, source, hard=True, is_repo=is_repo
            )
        else:
            projects = repo.install_from_directory(
                template, source, hard=True, is_repo=is_repo
            )
    except ProjectError as fnfe:
        console.error(str(fnfe))
    except FileExistsError as fee:
        console.error(str(fee))
    except shutil.Error:
        console.error(
            f"Could not install template {Fore.CYAN}{template}{Style.RESET_ALL}",
            echo=ctx.fail,
        )
    else:
        if type(projects) is type([]):
            for project in projects:
                console.success(
                    f"Installed template {Style.BRIGHT}{project.name}{Style.RESET_ALL}"
                )
            console.indent(
                f"Use with {Fore.MAGENTA}boil use <template_name>{Style.RESET_ALL}"
            )
        else:
            console.success(
                f"Installed template {Style.BRIGHT}{projects.name}{Style.RESET_ALL}"
            )
            console.indent(
                f"Use with {Fore.MAGENTA}boil use {projects.name}{Style.RESET_ALL}"
            )


@boil.command()
@click.option("-f", "--force", is_flag=True)
@click.argument("template")
@pass_tpldir
def uninstall(TPLDIR, force, template):
    repo = Repository(TPLDIR)

    if repo.is_installed(template):
        rm = force
        if not force:
            rm = console.question(
                f"Do you really want to uninstall template {Fore.CYAN}{template}{Style.RESET_ALL}",
                color=Fore.YELLOW,
                echo=click.confirm,
            )
        if rm:
            try:
                repo.uninstall(template)
                console.success(
                    f"Removed template {Style.BRIGHT}{template}{Style.RESET_ALL}"
                )
            except:
                console.error(
                    f"Error while uninstalling template {Fore.CYAN}{template}{Style.RESET_ALL}"
                )
                console.line(
                    f"You might need to manually delete the template directory at"
                )
                console.line(f"{Style.BRIGHT}{repo.root}{Style.RESET_ALL}")
    else:
        console.warn(f"Template {Fore.CYAN}{template}{Style.RESET_ALL} does not exist")


@boil.command()
@click.argument("template")
@click.pass_context
def update(ctx, template):
    """
    Update TEMPLATE from the source it was first installed from.
    """
    cfg = ctx.obj

    repo = Repository(cfg["TPLDIR"])

    if not repo.is_installed(template):
        console.error(
            f"Template {Fore.CYAN}{template}{Style.RESET_ALL} does not exist."
        )
        ctx.exit(2)

    project = repo.get_project(template)
    project.setup(load_project=True)
    try:
        project.update()
    except ProjectFileNotFoundError as pe:
        console.error(str(pe))
        console.indent(
            f"To update templates make sure to install with {Fore.MAGENTA}boil install{Style.RESET_ALL}."
        )
        ctx.abort()
    except ProjectError as pe:
        console.error(str(pe))
        ctx.abort()
    else:
        if project.meta["source_type"] == "github":
            console.success(
                f"Updated template {Fore.CYAN}{template}{Style.RESET_ALL} from GitHub."
            )
        else:
            console.success(
                f"Updated template {Fore.CYAN}{template}{Style.RESET_ALL} from local filesystem."
            )


@boil.command()
@click.option(
    "--hard",
    is_flag=True,
    help="Force overwrite of existing output directory. If the directory OUT exists and is not empty, it will be deleted and newly created.",
)
@click.option(
    "-v",
    "--value",
    multiple=True,
    nargs=2,
    help="Sets a prefilled value for the template.",
)
@click.argument("template")
@click.argument(
    "out", default=".", type=click.Path(file_okay=False, dir_okay=True, writable=True)
)
@click.pass_context
def use(ctx, template, out, hard, value):
    """
    Generate a new project from TEMPLATE.

    If OUT is given and a directory, the template is created there.
    Otherwise the cwd is used.
    """
    cfg = ctx.obj

    # Check teamplate and read configuration
    project = Project(template, cfg["TPLDIR"])
    try:
        project.setup(load_project=True)
    except FileNotFoundError:
        console.warn(
            f"No valid template found for key {Fore.CYAN}{template}{Style.RESET_ALL}"
        )
        ctx.exit(1)

    # Prepare output directory
    if out == ".":
        out = Path.cwd()
    else:
        out = Path(out)

    if out.exists() and len(os.listdir(out)) > 0:
        # log_warn(f'The output directory exists and is not empty.')
        # answ = log_question(f'Do you want to [{Style.BRIGHT}O{Style.RESET_ALL}]verwrite, [{Style.BRIGHT}M{Style.RESET_ALL}]erge or [{Style.BRIGHT}A{Style.RESET_ALL}]bort [o/m/a]')
        if hard:
            shutil.rmtree(out)
            out.mkdir(parents=True)
            console.success(f"Cleared {Style.BRIGHT}{out}{Style.RESET_ALL}")
    elif not out.exists():
        out.mkdir(parents=True)
        console.success(f"Created {Style.BRIGHT}{out}{Style.RESET_ALL}")

    # Read user input (if necessary)
    if project.fields:
        # Prepare dict for prefilled values
        prefilled = cfg["prefilled"] if "prefilled" in cfg else dict()
        if value:
            for k, v in value:
                prefilled[k] = v

        project.fill(prefilled)

    for success, file_in, file_out in project.compile(out):
        if success:
            console.success(f"Created {Style.BRIGHT}{file_out}{Style.RESET_ALL}")
        else:
            console.warn(
                f"Skipped {Style.BRIGHT}{file_in}{Style.RESET_ALL} due to empty content"
            )

    console.success(
        f'Generated project template "{Fore.CYAN}{template}{Style.RESET_ALL}" in {Style.BRIGHT}{out}{Style.RESET_ALL}'
    )



# -*- coding: utf-8 -*-

import click
from colorama import Fore, Back, Style


import parboil.console as console


def field_default(key, project, default="", value=None):
    if value:
        console.info(f'Used prefilled value for "{Fore.MAGENTA}{key}{Style.RESET_ALL}"')
        return value
    else:
        if type(default) == list:
            return field_choice(key, project, value=value, choices=default)
        elif type(default) is bool:
            if default:
                return not console.question(
                    f'Do you want do disable "{Fore.MAGENTA}{key}{Style.RESET_ALL}"',
                    echo=click.confirm,
                )
            else:
                return console.question(
                    f'Do you want do enable "{Fore.MAGENTA}{key}{Style.RESET_ALL}"',
                    echo=click.confirm,
                )
        else:
            return console.question(
                f'Enter a value for "{Fore.MAGENTA}{key}{Style.RESET_ALL}"',
                default=default,
            )


def field_choice(key, project, default=1, value=None, choices=list()):
    if value:
        console.info(f'Used prefilled value for "{Fore.MAGENTA}{key}{Style.RESET_ALL}"')
        project.variables[f"{key}_index"] = value
        return choices[value]
    else:
        if len(choices) > 1:
            console.question(
                f'Chose a value for "{Fore.MAGENTA}{key}{Style.RESET_ALL}"',
                echo=click.echo,
            )
            for n, choice in enumerate(choices):
                console.indent(f'{Style.BRIGHT}{n+1}{Style.RESET_ALL} -  "{choice}"')
            n = click.prompt(
                console.indent(f"Select from 1..{len(choices)}", echo=None),
                default=default,
            )
            if n > len(choices):
                console.warn(f"{n} is not a valid choice. Using default.")
                n = default
        else:
            n = 1
        project.variables[f"{key}_index"] = n - 1
        return choices[n - 1]


def field_file_select(
    key, project, default=1, value=None, choices=list(), filename=None
):
    if value and value in choices:
        console.info(f'Used prefilled value for "{Fore.MAGENTA}{key}{Style.RESET_ALL}"')
    else:
        if len(choices) > 1:
            console.question(
                f'Chose a value for "{Fore.MAGENTA}{key}{Style.RESET_ALL}"',
                echo=click.echo,
            )
            for n, choice in enumerate(choices):
                console.indent(f'{Style.BRIGHT}{n+1}{Style.RESET_ALL} -  "{choice}"')
            n = click.prompt(
                console.indent(f"Select from 1..{len(choices)}", echo=None),
                default=default,
            )
            if n > len(choices):
                console.warn(f"{n} is not a valid choice. Using default.")
                n = default
            value = choices[n - 1]
        else:
            value = choices[0]

    project.templates.append(f"includes:{value}")
    if filename:
        project.files[value] = filename
    return value
